# coding: utf-8
from unittest import TestCase

from edgy.project.events import LoggingDispatcher
from edgy.project.feature import Feature
from edgy.project.file import NullFile


class FeatureTestCase(TestCase):
    feature_type = Feature

    def create_dispatcher(self):
        return LoggingDispatcher()

    def create_feature(self, feature_type=None, dispatcher = None):
        dispatcher = dispatcher or self.create_dispatcher()
        feature = (feature_type or self.feature_type)(dispatcher)
        feature.file_type = NullFile
        return feature, dispatcher




# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

from edgy.project.events import subscribe

from . import Feature, SUPPORT_PRIORITY


class SphinxFeature(Feature):
    @subscribe('edgy.project.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile

        makefile['SPHINX_OPTS'] = ''
        makefile['SPHINX_BUILD'] = '$(VIRTUAL_ENV)/bin/sphinx-build'
        makefile['SPHINX_SOURCEDIR'] = 'doc'
        makefile['SPHINX_BUILDDIR'] = '$(SPHINX_SOURCEDIR)/_build'

        makefile.add_target('doc', '''
            $(SPHINX_BUILD) -b html -D latex_paper_size=a4 $(SPHINX_OPTS) $(SPHINX_SOURCEDIR) $(SPHINX_BUILDDIR)/html
        ''', deps=('install-dev',), phony=True)


__feature__ = SphinxFeature



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

from edgy.project.events import subscribe

from . import Feature, SUPPORT_PRIORITY


class Flake8Feature(Feature):
    requires = {'python'}

    @subscribe('edgy.project.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile.add_target('lint', '''
            $(VIRTUAL_ENV)/bin/flake8 {name}
        '''.format(name=event.package_name), deps=('install',), phony=True)


__feature__ = Flake8Feature



from __future__ import absolute_import

import os

from blessings import Terminal
from edgy.project.events import attach_subscriptions
from edgy.project.file import File
from edgy.project.settings import DEFAULT_FEATURES
from edgy.project.util import format_file_content
from jinja2 import Environment, PackageLoader, Template

ABSOLUTE_PRIORITY = -100
HIGH_PRIORITY = -80
MEDIUM_PRIORITY = -60
LOW_PRIORITY = -60
SUPPORT_PRIORITY = -20
LAST_PRIORITY = 100

try:
    input = raw_input  # pylint: disable=raw_input-builtin
except NameError:
    input = input  # pylint: disable=input-builtin

term = Terminal()


class Feature(object):
    _jinja_environment = None

    requires = set()
    conflicts = set()

    file_type = staticmethod(File)

    def __init__(self, dispatcher):
        """

        :param LoggingDispatcher dispatcher:
        """
        self.dispatcher = dispatcher
        self.configure()
        attach_subscriptions(self, self.dispatcher)

    def configure(self):
        pass

    def file(self, *args, **kwargs):
        return self.file_type(self.dispatcher, *args, **kwargs)

    @property
    def __name__(self):
        return type(self).__name__

    @property
    def __shortname__(self):
        return self.__name__.replace('Feature', '').lower()

    @property
    def jinja(self):
        if type(self)._jinja_environment is None:
            type(self)._jinja_environment = Environment(
                loader=PackageLoader(__name__, 'template')
            )
        return type(self)._jinja_environment

    def _log_file(self, target, override, content=()):
        self.dispatcher.info(term.bold(term.red('W!') if override else term.green('W?')), target,
                             '({} bytes)'.format(len(content)))

    def render(self, template, context=None):
        context = context or {}
        os.path.join(os.path.dirname(__file__), 'template')
        return self.jinja.get_template(template).render(**(context or {}))

    def render_file(self, target, template, context=None, override=False):
        with self.file(target, override=override) as f:
            content = format_file_content(self.render(template, context))
            f.write(content)
            self._log_file(target, override, content)

    def render_file_inline(self, target, template_string, context=None, override=False):
        with self.file(target, override=override) as f:
            content = format_file_content(Template(template_string).render(**(context or {})))
            f.write(content)
            self._log_file(target, override, content)

    def render_empty_files(self, *targets, **kwargs):
        override = kwargs.pop('override', False)
        for target in targets:
            with self.file(target, override=override) as f:
                self._log_file(target, override)


class ProjectInitializer(Feature):
    def execute(self):
        context = {}

        context['name'] = input('Name: ')
        context['description'] = input('Description: ')
        context['license'] = input('License [Apache License, Version 2.0]: ').strip() or 'Apache License, Version 2.0'

        context['url'] = ''
        context['download_url'] = ''

        context['author'] = ''
        context['author_email'] = ''

        context['features'] = DEFAULT_FEATURES
        context['install_requires'] = [

        ]
        context['extras_require'] = {
            'dev': [
                'coverage >=4.0,<4.2',
                'mock >=2.0,<2.1',
                'nose >=1.3,<1.4',
                'pylint >=1.6,<1.7',
                'pytest >=2.9,<2.10',
                'pytest-cov >=2.3,<2.4',
                'sphinx',
                'sphinx_rtd_theme',
            ]
        }
        context['entry_points'] = {}

        self.render_file('Projectfile', 'Projectfile.j2', context, override=True)



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import os

from edgy.project.events import subscribe

from . import ABSOLUTE_PRIORITY, Feature


class GitFeature(Feature):
    @subscribe('edgy.project.on_start', priority=ABSOLUTE_PRIORITY)
    def on_start(self, event):
        if not os.path.exists('.git'):
            self.dispatcher.info('git', 'Creating git repository...')
            os.system('git init')
            os.system('git add Projectfile')
            os.system('git commit -m "initial commit"')

    @subscribe('edgy.project.on_end')
    def on_end(self, event):
        self.render_file_inline('.gitignore', '''
            *.egg-info
            *.pyc
            *.swp
            /.cache
            /.coverage
            /.idea
            /.python*-*
            /build
            /dist
            /htmlcov
            /pylint.html
        ''', event.variables)

    @subscribe('edgy.project.on_file_closed')
    def on_file_change(self, event):
        os.system('git add {}'.format(event.filename))


__feature__ = GitFeature



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

from edgy.project.events import ProjectEvent
from edgy.project.feature.make import Makefile, MakefileEvent
from edgy.project.feature.tornado import TornadoFeature
from edgy.project.testing import FeatureTestCase

PACKAGE_NAME = 'foo.bar'


class TestTornadoFeature(FeatureTestCase):
    feature_type = TornadoFeature

    def test_configure(self):
        feature, dispatcher = self.create_feature()
        listeners = dispatcher.get_listeners()

        assert feature.on_start in listeners['edgy.project.on_start']
        assert feature.on_make_generate in listeners['edgy.project.feature.make.on_generate']

    def test_on_make_generate(self):
        feature, dispatcher = self.create_feature()
        makefile = Makefile()
        event = MakefileEvent(PACKAGE_NAME, makefile)

        # actual call
        feature.on_make_generate(event)

        # check makefile targets
        targets = dict(makefile.targets)
        assert {'serve', 'serve-wsgi'} == set(targets)

    def test_on_start(self):
        feature, dispatcher = self.create_feature()
        event = ProjectEvent(setup={'name': PACKAGE_NAME})
        feature.on_start(event)



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import contextlib
import os

from edgy.event import Event


class ShellEvent(Event):
    def __init__(self, command):
        super(ShellEvent, self).__init__()

class ShellWrapper(object):
    def __call__(self, *args, **kwargs):
        os.system(*args, **kwargs)

@contextlib.contextmanager
def Shell(dispatcher):
    event = ShellEvent()

    event = dispatcher.dispatch('edgy.project.on_command', event)
    yield f
    event.file = None






# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import os

from edgy.project.events import subscribe

from . import Feature, SUPPORT_PRIORITY


class TornadoFeature(Feature):
    requires = {'python'}
    conflicts = {'django', 'flask'}

    @subscribe('edgy.project.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile

        makefile.add_target('serve', '''
            $(VIRTUAL_ENV)/bin/python -m {name}.app
        '''.format(name=event.package_name), deps=('install',), phony=True)

        makefile.add_target('serve-wsgi', '''
            $(VIRTUAL_ENV)/bin/python -m {name}.wsgi
        '''.format(name=event.package_name), deps=('install',), phony=True)

    @subscribe('edgy.project.on_start', priority=SUPPORT_PRIORITY)
    def on_start(self, event):
        package_path = event.setup['name'].replace('.', os.sep)

        for file in 'app', 'handlers', 'settings', 'wsgi':
            self.render_file(
                os.path.join(package_path, file + '.py'),
                os.path.join('tornado', file + '.py.j2')
            )


__feature__ = TornadoFeature



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import pytest
from edgy.event import Event
from edgy.project.events import ProjectEvent
from edgy.project.feature.git import GitFeature
from edgy.project.testing import FeatureTestCase
from mock import patch

PACKAGE_NAME = 'foo.bar'


class TestGitFeature(FeatureTestCase):
    feature_type = GitFeature

    def test_configure(self):
        feature, dispatcher = self.create_feature()
        listeners = dispatcher.get_listeners()

        assert feature.on_start in listeners['edgy.project.on_start']
        assert feature.on_end in listeners['edgy.project.on_end']
        assert feature.on_file_change in listeners['edgy.project.on_file_closed']

    def test_on_start(self):
        feature, dispatcher = self.create_feature()

        with patch('os.path.exists', return_value=False):
            commands = list()
            with patch('os.system', side_effect=commands.append) as os_system:
                feature.on_start(Event())
                assert commands == ['git init', 'git add Projectfile', 'git commit -m "initial commit"']

        with patch('os.path.exists', return_value=True):
            commands = list()
            with patch('os.system', side_effect=commands.append) as os_system:
                feature.on_start(Event())
                assert commands == []

    def test_on_end(self):
        feature, dispatcher = self.create_feature()

        commands = list()
        with patch('edgy.project.file.FileEvent') as fe, \
                patch('os.system', side_effect=commands.append) as os_system \
                :
            feature.on_end(ProjectEvent(setup={'name': PACKAGE_NAME}))
            print(commands)
            print(fe.call_args_list)

    @pytest.skip
    def test_on_file_change(self):
        self.fail()



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

DEFAULT_FEATURES = {
    'git',
    'make',
    'pytest',
    'python',
    'pylint',
}

DEFAULT_FILES = {
    'requirements',
    'requirements-dev',
    'classifiers',
    'version',
}



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

from edgy.project.events import subscribe

from . import Feature, SUPPORT_PRIORITY


class NosetestsFeature(Feature):
    requires = {'python'}
    conflicts = {'pytest'}

    @subscribe('edgy.project.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile.add_target('test', '''
            $(VIRTUAL_ENV)/bin/nosetests -q --with-doctest --with-coverage --cover-package={name}
        '''.format(name=event.package_name), deps=('install-dev',), phony=True)

    @subscribe('edgy.project.feature.python.on_generate', priority=SUPPORT_PRIORITY)
    def on_python_generate(self, event):
        event.files['requirements'] = '\n'.join((event.files['requirements'], 'nose >=1.3,<1.4', 'coverage >=3.7,<3.8'))


__feature__ = NosetestsFeature



from __future__ import absolute_import

from .util import format_file_content

class Config(object):
    def __init__(self):
        pass

def read_configuration(dispatcher, filename, variables, features, files, setup):
    with open(filename) as f:
        code = compile(f.read(), filename, 'exec')
    ctx = {
        'listen': dispatcher.listen
    }
    exec(code, ctx)

    for k in variables.keys():
        if k in ctx:
            variables[k] = ctx[k]

    features = (features | set(ctx.pop('enable_features', ()))) - set(ctx.pop('disable_features', ()))

    for k in files.keys():
        if k in ctx:
            files[k] = format_file_content(ctx[k])

    for k in setup.keys():
        if k in ctx:
            setup[k] = ctx[k]
        if setup[k] is None:
            raise ValueError('You must provide a value for the setup entry "{}" in your Projectfile.'.format(k))

    return variables, features, files, setup



#!/usr/bin/env python
# coding: utf-8

from __future__ import absolute_import, print_function

import argparse
import logging
import os
import sys
from collections import OrderedDict

import tornado.log
from blessings import Terminal
from edgy.event import EventDispatcher
from edgy.project.config import read_configuration
from edgy.project.events import LoggingDispatcher, ProjectEvent
from edgy.project.feature import ProjectInitializer
from edgy.project.settings import DEFAULT_FEATURES, DEFAULT_FILES

# Globals
logger = logging.getLogger()
tornado.log.enable_pretty_logging(logger=logger)
t = Terminal()


def _read_configuration(dispatcher, config_filename):
    """
    Prepare the python context and delegate to the real configuration reader (see config.py)

    :param EventDispatcher dispatcher:
    :return tuple: (variables, features, files, setup)
    """
    if not os.path.exists(config_filename):
        raise IOError('Could not find project description file (looked in {})'.format(config_filename))

    variables = OrderedDict((
        ('virtual_env', '.virtualenv-$(PYTHON_BASENAME)',),
    ))

    files = {filename: '' for filename in DEFAULT_FILES}

    setup = OrderedDict((
        ('name', None,),
        ('description', None,),
        ('license', None,),
        ('url', 'http://example.com/',),
        ('download_url', 'http://example.com/',),
        ('extras_require', {},),
        ('install_requires', [],),
        ('entry_points', {},),
    ))

    features = set(DEFAULT_FEATURES)

    return read_configuration(dispatcher, config_filename, variables, features, files, setup)


def main(args=None):
    parser = argparse.ArgumentParser()

    parser.add_argument('--config', '-c', default='Projectfile')
    parser.add_argument('--verbose', '-v', action='store_true', default=False)

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    parser_init = subparsers.add_parser('init', help='Initialize a new project.')
    parser_init.set_defaults(handler=handle_init)

    parser_update = subparsers.add_parser('update', aliases=('up',), help='Update current project.')
    parser_update.set_defaults(handler=handle_update)

    options = parser.parse_args(args or sys.argv[1:])
    if options.verbose:
        logger.setLevel(logging.DEBUG)

    return options.handler(
        os.path.join(os.getcwd(), options.config)
    )


def handle_init(config_filename):
    if os.path.exists(config_filename):
        raise IOError(
            'No config should be present in current directory to initialize (found {})'.format(config_filename))

    # Fast and dirty implementation
    # TODO
    # - input validation
    # - getting input from env/git conf (author...),
    # - dispatching something in selected features so maybe they can suggest deps
    # - deps selection
    # - ...
    dispatcher = LoggingDispatcher()
    initializer = ProjectInitializer(dispatcher)
    initializer.execute()
    return handle_update(config_filename)


def handle_update(config_filename):
    dispatcher = LoggingDispatcher()

    variables, features, files, setup = _read_configuration(dispatcher, config_filename)

    feature_instances = {}
    logger.info('Updating {} with {} features'.format(
        t.bold(setup['name']),
        ', '.join(t.bold(t.green(feature_name)) for feature_name in sorted(features))))

    sorted_features = sorted(features)
    for feature_name in sorted_features:
        logger.debug('Initializing feature {}...'.format(t.bold(t.green(feature_name))))
        try:
            feature = __import__('edgy.project.feature.' + feature_name, fromlist=('__feature__',)).__feature__
        except (ImportError, AttributeError,) as e:
            logger.exception('Feature "{}" not found.'.format(feature_name))

        if feature:
            feature_instances[feature_name] = feature(dispatcher)

            for req in feature_instances[feature_name].requires:
                if not req in sorted_features:
                    raise RuntimeError('Unmet dependency: {} requires {}.'.format(feature_name, req))

            for con in feature_instances[feature_name].conflicts:
                if con in sorted_features:
                    raise RuntimeError('Conflicting dependency: {} conflicts with {}.'.format(con, feature_name))
        else:
            raise RuntimeError('Required feature {} not found.'.format(feature_name))

    event = ProjectEvent()
    event.variables, event.files, event.setup = variables, files, setup

    # todo: add listener dump list in debug/verbose mode ?

    event = dispatcher.dispatch('edgy.project.on_start', event)

    dispatcher.dispatch('edgy.project.on_end', event)

    logger.info(u'Done.')


if __name__ == '__main__':
    main()



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

from edgy.project.events import subscribe

from . import Feature, SUPPORT_PRIORITY


class PylintFeature(Feature):
    requires = ['python']

    @subscribe('edgy.project.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile.add_target('lint', '''
            $(VIRTUAL_ENV)/bin/pylint --py3k {name} -f html > pylint.html
        '''.format(name=event.package_name), deps=('install-dev',), phony=True)

    @subscribe('edgy.project.feature.python.on_generate', priority=SUPPORT_PRIORITY)
    def on_python_generate(self, event):
        event.files['requirements'] = '\n'.join((event.files['requirements'], 'pylint >=1.4,<1.5',))


__feature__ = PylintFeature



# coding: utf-8

from __future__ import absolute_import, unicode_literals

import os
from getpass import getuser

from edgy.project.events import subscribe

from . import ABSOLUTE_PRIORITY, Feature


class PythonFeature(Feature):
    """
    Basic python support for your project. Adds the requirements/requirements-dev logic, virtualenv support, setup.py
    generation, a few metadata files used by setuptools...

    """

    @subscribe('edgy.project.feature.make.on_generate', priority=ABSOLUTE_PRIORITY)
    def on_make_generate(self, event):
        """
        **Environment variables**

        - ``PYTHON``
        - ``PYTHON_BASENAME``
        - ``PYTHON_REQUIREMENTS_FILE``
        - ``PYTHON_REQUIREMENTS_DEV_FILE``
        - ``PIP`` (should it be renamed to PYTHON_PIP to match the naming pattern?)

        **Make targets**

        - ``install``
        - ``install-dev``
        - ``$(VIRTUAL_ENV)``: either a local brand new python virtualenv or your currently activated virtualenv.

        :param MakefileEvent event:

        """
        # Python related environment
        event.makefile.updateleft(
            ('PYTHON', '$(shell which python)',),
            ('PYTHON_BASENAME', '$(shell basename $(PYTHON))',),
            ('PYTHON_REQUIREMENTS_FILE', 'requirements.txt',),
            ('PYTHON_REQUIREMENTS_DEV_FILE', 'requirements-dev.txt',),
        )

        # Package manager
        event.makefile['PIP'] = '$(VIRTUAL_ENV)/bin/pip'

        # Virtualenv, with external virtual env support.
        event.makefile.add_target('$(VIRTUAL_ENV)', '''
            virtualenv -p $(PYTHON) $(VIRTUAL_ENV)
            $(PIP) install -U pip\>=8.1.2,\<9 wheel\>=0.29,\<1.0
            ln -fs $(VIRTUAL_ENV)/bin/activate activate-$(PYTHON_BASENAME)
        ''', doc='''
            Setup the local virtualenv, or use the one provided by the current environment.
        ''')
        event.makefile.set_deps('install', ('$(VIRTUAL_ENV)',))
        event.makefile.get_target('install').install = [
            '$(PIP) install -Ur $(PYTHON_REQUIREMENTS_FILE)'
        ]

        event.makefile.set_deps('install-dev', ('$(VIRTUAL_ENV)',))
        event.makefile.get_target('install-dev').install = [
            '$(PIP) install -Ur $(PYTHON_REQUIREMENTS_DEV_FILE)'
        ]

    @subscribe('edgy.project.on_start', priority=ABSOLUTE_PRIORITY)
    def on_start(self, event):
        """
        **Events**

        - ``edgy.project.feature.python.on_generate`` (with the same ``ProjectEvent`` we got, todo: why not deprecate
          it in favor of higher priority edgy.project.on_start?)

        **Files**

        - ``<yourpackage>/__init__.py``
        - ``MANIFEST.in``
        - ``README.rst``
        - ``classifiers.txt``
        - ``requirements-dev.txt``
        - ``requirements.txt``
        - ``setup.cfg``
        - ``setup.py`` (overwritten)
        - ``version.txt``

        :param ProjectEvent event:
        """
        self.dispatcher.dispatch(__name__ + '.on_generate', event)

        # Some metadata that will prove useful.
        self.render_empty_files('classifiers.txt', 'version.txt', 'README.rst')
        self.render_file_inline('MANIFEST.in', 'include *.txt')
        self.render_file_inline('setup.cfg', '''
                    [metadata]
                    description-file = README.rst
                ''')

        # If requirements files are missing, let's create em with reasonable defaults.
        self.render_file_inline('requirements.txt', '-e .')
        self.render_file_inline('requirements-dev.txt', '-e .[dev]')

        # Explode package name so we know which python packages are namespaced and
        # which are not.
        package_bits = event.setup['name'].split('.')
        if len(package_bits) > 1:
            event.setup['namespace_packages'] = []
            for i in range(1, len(package_bits)):
                # TODO convert to string type (six?) depending on python version
                event.setup['namespace_packages'].append('.'.join(package_bits[0:i]))

        for i in range(1, len(package_bits) + 1):
            _bits = package_bits[0:i]
            package_dir = os.path.join(*_bits)
            if not os.path.exists(package_dir):
                os.makedirs(package_dir)

            package_init_file = os.path.join(package_dir, '__init__.py')
            if not os.path.exists(package_init_file):
                self.render_file(package_init_file, 'python/package_init.py.j2',
                                 {'is_namespace': i < len(package_bits)})

        context = {
            'url': event.setup.pop('url', 'http://example.com/'),
            'download_url': event.setup.pop('download_url', 'http://example.com/'),
        }

        for k, v in context.items():
            context[k] = context[k].format(
                name=event.setup['name'],
                user=getuser(),
                version='{version}',
            )

        context.update({
            'setup': event.setup,
            'extras_require': event.setup.pop('extras_require', {}),
            'install_require': event.setup.pop('install_require', {}),
            'entry_points': event.setup.pop('entry_points', {}),
        })

        # Render (with overwriting) the allmighty setup.py
        self.render_file('setup.py', 'python/setup.py.j2', context, override=True)


__feature__ = PythonFeature



from __future__ import absolute_import, unicode_literals

import contextlib
import os

from edgy.event import Event


class FileEvent(Event):
    def __init__(self, filename, override):
        super(FileEvent, self).__init__()
        self.filename = filename
        self.override = override


@contextlib.contextmanager
def File(dispatcher, name, override=False):
    event = FileEvent(name, override)

    if event.override or not os.path.exists(event.filename):
        with open(event.filename, 'w+') as f:
            event.file = f
            event = dispatcher.dispatch('edgy.project.on_file_opened', event)
            yield f
            event.file = None

        dispatcher.dispatch('edgy.project.on_file_closed', event)
    else:
        yield open('/dev/null', 'w')


@contextlib.contextmanager
def NullFile(dispatcher, name, override=False):
    event = FileEvent(name, override)

    if event.override or not os.path.exists(event.filename):
        with open('/dev/null', 'w') as f:
            event.file = f
            event = dispatcher.dispatch('edgy.project.on_file_opened', event)
            yield f
            event.file = None

        dispatcher.dispatch('edgy.project.on_file_closed', event)
    else:
        yield open('/dev/null', 'w')



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import pytest
from edgy.project.events import ProjectEvent
from edgy.project.feature.make import MakeFeature, Makefile, MakefileEvent
from edgy.project.feature.python import PythonFeature
from edgy.project.testing import FeatureTestCase

PACKAGE_NAME = 'foo.bar'


class TestPythonFeature(FeatureTestCase):
    feature_type = PythonFeature

    def test_configure(self):
        feature, dispatcher = self.create_feature()
        listeners = dispatcher.get_listeners()

        assert feature.on_start in listeners['edgy.project.on_start']
        assert feature.on_make_generate in listeners['edgy.project.feature.make.on_generate']

    def test_on_make_generate(self):
        python_feature, dispatcher = self.create_feature()

        with pytest.raises(KeyError):
            # This should fail, as the install/install-dev targets are defined in MakeFeature
            python_feature.on_make_generate(
                MakefileEvent(PACKAGE_NAME, Makefile())
            )

        make_feature, dispatcher = self.create_feature(feature_type=MakeFeature)
        self.create_feature(dispatcher=dispatcher)
        make_feature.on_start(
            ProjectEvent(setup={'name': PACKAGE_NAME})
        )

        assert sorted(dict(make_feature.makefile.targets).keys()) == [
            '$(VIRTUAL_ENV)',
            'clean',
            'install',
            'install-dev',
        ]

    def test_on_start(self):
        feature, dispatcher = self.create_feature()
        feature.on_start(
            ProjectEvent(setup={'name': PACKAGE_NAME})
        )



# This file is autogenerated by edgy.project code generator.
# All changes will be overwritten.

from setuptools import setup, find_packages

tolines = lambda c: list(filter(None, map(lambda s: s.strip(), c.split('\n'))))

def read(filename, flt=None):
    with open(filename) as f:
        content = f.read().strip()
        return flt(content) if callable(flt) else content

version = read('version.txt')

setup(
    name = 'edgy.project',
    description = 'Strongly opinionated python project management.',
    license = 'Apache License, Version 2.0',
    install_requires = ['blessings >=1.6,<1.7',
 'edgy.event >=0.1,<0.2',
 'jinja2 >=2.8,<3.0',
 'six',
 'tornado >=4.4,<4.5'],
    namespace_packages = ['edgy'],
    version = version,
    long_description = read('README.rst'),
    classifiers = read('classifiers.txt', tolines),
    packages = find_packages(exclude=['ez_setup', 'example', 'test']),
    include_package_data = True,
    extras_require = {'dev': ['coverage >=4.0,<4.2',
         'mock >=2.0,<2.1',
         'nose >=1.3,<1.4',
         'pylint >=1.6,<1.7',
         'pytest >=2.9,<2.10',
         'pytest-cov >=2.3,<2.4',
         'sphinx',
         'sphinx_rtd_theme']},
    entry_points = {'console_scripts': ['edgy-project=edgy.project.__main__:main']},
    url = 'https://github.com/python-edgy/project',
    download_url = 'https://github.com/python-edgy/project/tarball/{version}'.format(version=version),
)



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

import os

from edgy.project.events import subscribe

from . import Feature, SUPPORT_PRIORITY


class PytestFeature(Feature):
    requires = {'python'}
    conflicts = {'nosetests'}

    @subscribe('edgy.project.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile['PYTEST_OPTIONS'] = '--capture=no --cov={path} --cov-report html'.format(
            path=event.package_name.replace('.', os.sep)
        )
        makefile.add_target('test', '''
            $(VIRTUAL_ENV)/bin/py.test $(PYTEST_OPTIONS) tests
        ''', deps=('install-dev',), phony=True)

    @subscribe('edgy.project.on_start', priority=SUPPORT_PRIORITY)
    def on_start(self, event):
        tests_dir = os.path.join('tests', *event.setup['name'].split('.'))
        if not os.path.exists(tests_dir):
            os.makedirs(tests_dir)

        tests_init_file = os.path.join(tests_dir, '__init__.py')

        if not os.path.exists(tests_init_file):
            self.render_file(tests_init_file, 'python/package_init.py.j2', {'is_namespace': False})


__feature__ = PytestFeature
